import enum


class ConjuntoType(enum.Enum):
    PRODUTOS = 0
    UNIDADES = 1
    COMBUSTIVEIS = 2
    SERVICOS = 3
    CLASSIFICACOESPARTICIPANTES = 4
    FICHAS = 5
    CLIENTES = 6
    FORNECEDORES = 7
    TRANSPORTADORAS = 8
    VENDEDORES = 9
    SERVICOSDECATALOGOS = 1
    MODELOSCONTRATOS = 1
    TECNICOS = 1
    RUBRICAS = 1
    REPRESENTANTESCOMERCIAIS = 1
    REPRESENTANTESTECNICOS = 1
    PROSPECTS = 1
